"""
Problem
=======

A problem contains the objective as well as all information like prior
describing the problem to be solved.

"""

from .base import Problem
from .hierarchical import HierarchicalProblem
